""" QUInt and QInt class definition """

from math import pi
from typing import Literal, Optional, Type, TypeVar

from .qpu import QPU
from .qubit import Qubit
from .rng import _rng

QIntType = TypeVar("QIntType", bound="QUInt")


class QUInt:

  def __init__(self: QIntType,
               num_qubits: int,
               qpu: Optional[QPU] = None) -> None:
    self.num_qubits = num_qubits

    if qpu is None:
      qpu = QPU(num_qubits)
    self.qpu = qpu
    self.start_qubit = qpu._next_qubit
    self.qpu._reserve(num_qubits)

  def __getitem__(self, qubit: int) -> Qubit:
    return Qubit(self, qubit)

  def __setitem__(self, qubit: int, value: Literal[1, 0]) -> None:
    # TODO(dyordan1): Is this really the most reliable way to "force" a qubit?
    if value != 1 and value != 0:
      raise Exception("Can only set qubits to 0 or 1")

    # sanity check if we can set this bit
    probability = self.qpu.GetProbability(self.start_qubit + qubit)
    if (probability == 0 and value == 1) or (probability == 1 and value == 0):
      if not self.qpu.IsClassicalBit(self.start_qubit + qubit, 1.e-13):
        raise Exception("Cannot set entangled qubit")

      # Invert gate if incompatible *and* classical
      self.qpu.ApplyPauliX(self.start_qubit + qubit)

    self.qpu.CollapseQubit(self.start_qubit + qubit, bool(value))
    self.qpu.Normalize()

  @classmethod
  def zeros(cls: Type[QIntType],
            size: int,
            *,
            qpu: Optional[QPU] = None) -> QIntType:
    return cls(size, qpu=qpu)

  @classmethod
  def ones(cls: Type[QIntType],
           size: int,
           *,
           qpu: Optional[QPU] = None) -> QIntType:
    return cls.zeros(size, qpu=qpu).negate()

  @classmethod
  def pluses(cls: Type[QIntType],
             size: int,
             *,
             qpu: Optional[QPU] = None) -> QIntType:
    return cls.zeros(size, qpu=qpu).hadamard()

  @classmethod
  def minuses(cls: Type[QIntType],
              size: int,
              *,
              qpu: Optional[QPU] = None) -> QIntType:
    return cls.ones(size, qpu=qpu).hadamard()

  @classmethod
  def lefts(cls: Type[QIntType],
            size: int,
            *,
            qpu: Optional[QPU] = None) -> QIntType:
    return cls.zeros(size, qpu=qpu).sqrt_not().hadamard()

  @classmethod
  def rights(cls: Type[QIntType],
             size: int,
             *,
             qpu: Optional[QPU] = None) -> QIntType:
    return cls.ones(size, qpu=qpu).sqrt_not().hadamard()

  @classmethod
  def fully_entangled(cls: Type[QIntType],
                      size: int,
                      *,
                      qpu: Optional[QPU] = None) -> QIntType:
    q = cls.zeros(size, qpu=qpu).hadamard(0)

    for i in range(q.num_qubits - 1):
      q.qpu.ApplyCPauliX(q.start_qubit + i, q.start_qubit + i + 1)

    return q

  @classmethod
  def value(cls: Type[QIntType],
            size: int,
            value: int,
            *,
            qpu: Optional[QPU] = None) -> QIntType:
    return cls.zeros(size, qpu=qpu).set(value)

  def negate(self: QIntType, qubit: Optional[int] = None) -> QIntType:
    if qubit is not None:
      self.qpu.ApplyPauliX(self.start_qubit + qubit)
      return self

    for i in range(self.num_qubits):
      self.negate(i)

    return self

  def c_negate(self: QIntType,
               *,
               on: Qubit,
               qubit: Optional[int] = None) -> QIntType:
    if self.qpu != on.quint.qpu:
      raise Exception("Cannot cross-compute between QPUs!")

    if qubit is not None:
      self.qpu.ApplyCPauliX(on.quint.start_qubit + on.qubit,
                            self.start_qubit + qubit)
      return self

    for i in range(self.num_qubits):
      self.c_negate(on=on, qubit=i)

    return self

  def sqrt_not(self: QIntType, qubit: Optional[int] = None) -> QIntType:
    if qubit is not None:
      self.qpu.ApplyPauliSqrtX(self.start_qubit + qubit)
      return self

    for i in range(self.num_qubits):
      self.sqrt_not(i)

    return self

  def hadamard(self: QIntType, qubit: Optional[int] = None) -> QIntType:
    if qubit is not None:
      self.qpu.ApplyHadamard(self.start_qubit + qubit)
      return self

    for i in range(self.num_qubits):
      self.hadamard(i)

    return self

  def phase(self: QIntType,
            *,
            qubit: Optional[int] = None,
            angle: float = -pi / 2) -> QIntType:
    if qubit is not None:
      self.qpu.ApplyRotationZ(self.start_qubit + qubit, angle)
      return self

    for i in range(self.num_qubits):
      self.phase(qubit=i, angle=angle)

    return self

  def c_phase(self: QIntType,
              *,
              on: Qubit,
              qubit: Optional[int] = None,
              angle: float = -pi / 2) -> QIntType:
    if self.qpu != on.quint.qpu:
      raise Exception("Cannot cross-compute between QPUs!")

    if qubit is not None:
      self.qpu.ApplyCRotationZ(on.quint.start_qubit + on.qubit,
                               self.start_qubit + qubit, angle)
      return self

    for i in range(self.num_qubits):
      self.c_phase(on=on, qubit=i, angle=angle)

    return self

  def swap(self: QIntType,
           other: QIntType,
           qubit: Optional[int] = None,
           other_qubit: Optional[int] = None) -> QIntType:
    if self.qpu != other.qpu:
      raise Exception("Cannot cross-compute between QPUs!")

    if qubit is not None:
      if other_qubit is None:
        other_qubit = qubit

      self.qpu.ApplySwap(self.start_qubit + qubit,
                         other.start_qubit + other_qubit)
      return self

    if self.num_qubits != other.num_qubits:
      raise Exception("Cannot swap Q(U)Int of different sizes!")

    for i in range(self.num_qubits):
      self.swap(other, i)

    return self

  def c_swap(self: QIntType,
             other: QIntType,
             *,
             on: Qubit,
             qubit: Optional[int] = None,
             other_qubit: Optional[int] = None) -> QIntType:
    if self.qpu != other.qpu or self.qpu != on.quint.qpu:
      raise Exception("Cannot cross-compute between QPUs!")

    if qubit is not None:
      if other_qubit is None:
        other_qubit = qubit

      self.qpu.ApplyToffoli(on.quint.start_qubit + on.qubit,
                            self.start_qubit + qubit,
                            other.start_qubit + other_qubit)
      self.qpu.ApplyToffoli(on.quint.start_qubit + on.qubit,
                            other.start_qubit + other_qubit,
                            self.start_qubit + qubit)
      self.qpu.ApplyToffoli(on.quint.start_qubit + on.qubit,
                            self.start_qubit + qubit,
                            other.start_qubit + other_qubit)
      return self

    if self.num_qubits != other.num_qubits:
      raise Exception("Cannot swap Q(U)Int of different sizes!")

    for i in range(self.num_qubits):
      self.c_swap(other, on=on, qubit=i)

    return self

  def swap_test(self: QIntType,
                other: QIntType,
                *,
                on: Qubit,
                qubit: Optional[int] = None,
                other_qubit: Optional[int] = None) -> QIntType:
    if self.qpu != other.qpu or self.qpu != on.quint.qpu:
      raise Exception("Cannot cross-compute between QPUs!")

    # Trample the ancilla for a swaptest
    on.quint[on.qubit] = 0
    self.c_swap(other, on=on.hadamard(), qubit=qubit, other_qubit=other_qubit)
    return on.hadamard().negate().measure()

  def set(self, value: int) -> QIntType:
    if value < 0 or value > 2**self.num_qubits:
      raise Exception(f"Value to set for QUInt out of bounds: {value}")
    mask = 1
    for i in range(self.num_qubits):
      self[i] = 1 if value & mask else 0
      mask <<= 1

    return self

  def measure(self: QIntType, qubit: Optional[int] = None) -> int:
    if qubit is not None:
      probability = self.qpu.GetProbability(self.start_qubit + qubit)
      r = _rng.GetUniformRandomNumbers(1, 0., 1., "pool")[0]
      if r < probability:
        self.qpu.CollapseQubit(self.start_qubit + qubit, True)
        self.qpu.Normalize()
        return 1

      self.qpu.CollapseQubit(self.start_qubit + qubit, False)
      self.qpu.Normalize()
      return 0

    val = 0
    for i in range(self.num_qubits - 1, -1, -1):
      val <<= 1
      val |= self.measure(i)

    return val

  # Aliases
  h = hadamard
  dburby = hadamard
  entangle = c_negate


class QInt(QUInt):

  def measure(self: QIntType, qubit: Optional[int] = None) -> int:
    val = super().measure(qubit)
    if qubit is not None:
      return val

    if val < (1 << self.num_qubits - 1):
      return val
    val ^= 2**(self.num_qubits) - 1
    return -(val + 1)

  def set(self, value: int) -> QIntType:
    if value < -(1 << self.num_qubits - 1) or value >= (
        1 << self.num_qubits - 1):
      raise Exception(f"Value to set for QInt out of bounds: {value}")

    if value < 0:
      value ^= 2**(self.num_qubits) - 1
      value = -(value + 1)

    return super().set(value)



""" QPU class definition """

import cmath
from dataclasses import dataclass
import math
from typing import List

from .iqs import _iqs


@dataclass
class QPUState:
  value: int
  probability: float
  phase: float


@dataclass
class QPUStateCollection:
  states: List[QPUState]

  def __getitem__(self, i):
    return self.states[i]

  def __len__(self):
    return len(self.states)

  def debug_string(self):
    debug_vals = []
    known_phases = {
        5.759586: "11π/6",
        5.497787: "7π/4",
        5.235987: "5π/3",
        4.712388: "3π/2",
        4.188790: "4π/3",
        3.926990: "5π/4",
        3.665191: "7π/6",
        3.141592: "π",
        2.617993: "5π/6",
        2.356194: "3π/4",
        2.094395: "2π/3",
        1.570796: "π/2",
        1.047197: "π/3",
        0.785398: "π/4",
        0.523598: "π/6"
    }
    for state in self.states:
      debug = f"{state.value}: {round(state.probability, 4)*100}%"
      if abs(state.phase) > 0.000001:
        phase = state.phase
        for known_val, known_label in known_phases.items():
          if abs(abs(state.phase) - known_val) < 0.000001:
            phase = known_label
            # TODO(dyordan1): Do we need phase < 0? Some sources use [-pi; pi]
            # instead of [0; 2pi]
            if state.phase * known_val < 0:
              phase = "-" + phase
            break
        debug += f" ({phase})"
      debug_vals.append(debug)
    return "QPU states:\n" + "\n".join(debug_vals) + "\n"


class QPU(_iqs.QubitRegister):
  """ Utility class to represent a QPU that can be split up into Q(U)Ints

  Args:
    num_qubits: How many qubits to allocated to the QPU

  """

  def __init__(self, num_qubits: int) -> None:
    self.num_qubits = num_qubits
    self._next_qubit = 0
    super().__init__(num_qubits, "base", 0, 0)

  def _reserve(self, num_qubits: int):
    if self._next_qubit + num_qubits > self.num_qubits:
      raise Exception(
          "Overcommitted QPU! You tried to reserve more qubits than available!")

    self._next_qubit += num_qubits

  def states(self) -> List[QPUState]:
    self.Normalize()
    states = []
    for i in range(self.GlobalSize()):
      amplitude = self.GetGlobalAmplitude(i)
      magnitude = abs(amplitude)
      states.append(
          QPUState(value=i,
                   probability=magnitude * magnitude,
                   phase=cmath.phase(amplitude)))
    state_collection = QPUStateCollection(
        states=list(filter(lambda s: s.probability != 0, states)))
    return state_collection



#------------------------------------------------------------------------------
# Copyright (C) 2020 Intel Corporation 
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------


'''Tutorial on the use of Intel Quantum Simulator (IQS) for parallel noisy simulation.'''

# Import relevant libraries
import sys
sys.path.insert(0, '../build/lib')
import intelqs_py as iqs
import numpy as np

#################################################################################
# Setting the MPI environment
#################################################################################

# Initialize the MPI environment 
iqs.EnvInit()
num_procs = iqs.MPIEnvironment.GetSizeWorldComm()
#print("size world comm =", iqs.MPIEnvironment.GetSizeWorldComm(), flush=True)

# Utility function to print a message from the master process only.
def info(message):
    if iqs.MPIEnvironment.GetPoolRank()==0 and iqs.MPIEnvironment.IsUsefulRank():
        print(message, flush=True)

info("\nThis noisy simulation is thought to be run with MPI.")
info("To do so, please set the option '-DIqsMPI=ON' when calling CMake.")
info("However the code will execute also without MPI.\n")

#################################################################################

# In IQS the noise is introduced using the Stochastic Schroedinger Equation,
# implemented via noise gates. This means that a single (noisy) quantum circuit
# is substituted by an ensemble of many (ideal) circuits, each having
# additional stochastic gates and corresponding to a specific realization
# of the stochastic term of the equation.

# First of all, one needs to decide how many distinct states to simulate in the pool.
# In this example, we use a single rank per state.
num_pool_states = num_procs

# Partition the MPI environment into groups of processes. One group per pool state.
with_mpi_info = True
iqs.MPIEnvironment.UpdateStateComm(num_pool_states, with_mpi_info)

# Here it is unnecessary, but sometimes there are dummy MPI processes that do not perform any work.
# Finalize them and terminate the run.
if iqs.MPIEnvironment.IsUsefulRank()==False:
    iqs.EnvFinalize()
    exit()

# Second, one needs to decide how many states form the ensemble.
# Here we choose the smaller multiple of the num_pool_states grater or equal to 200.
min_num_ensemble_states = 200
num_ensemble_states = min_num_ensemble_states
if min_num_ensemble_states%num_pool_states != 0:
    num_ensemble_states += num_pool_states-(min_num_ensemble_states%num_pool_states)
if num_ensemble_states%num_pool_states != 0:
    print('ERROR: the ensemble does not contain a number of states multiple of the pool')

# IQS has functions that simplify some MPI instructions.
# However, it is important to keep trace of the current rank.
my_rank = iqs.MPIEnvironment.GetPoolRank()


#################################################################################
# Quantum state initialization and ideal circuit
#################################################################################

# The circuit consist of three single-qubit rotations per qubit, the first
# in the X basis, the second in the Y basis and the third in the Z basis.

# The noisy simulation depends on the gate parallelism. Here we consider the
# sequential application of the gates, one at a time. The order is:
# 1. X rotations are implemented, starting from qubit 0 and in increasing qubit order;
# 2. Y rotations are implemented, starting from qubit 0 and in increasing qubit order;
# 3. Z rotations are implemented, starting from qubit 0 and in increasing qubit order.

# To generate random numbers, IQS provides a wrapper around VSL random number generator.
# If MKL is not available, a standard MT19937 generator is used.
# We need to declare the (pseudo) random number generator...
rng = iqs.RandomNumberGenerator()
# ... and initialize its seed:
rng_seed = 77777
rng.SetSeedStreamPtrs( rng_seed )

# NOTE: the random number generator is able to generate three different kinds
#       of random numbers:
#       *local* --> different for each pool rank
#       *state* --> common to all ranks of the same state
#       *pool*  --> common to all ranks of the pool

# All angles of rotations are random. However once the circuit we want to simulate
# is determined, there is no stochasticity in the rotation angles across the ensemble.
# The rotation angles must therefore be given as *pool* random numbers.

num_qubits = 8
x_angles = rng.GetUniformRandomNumbers(num_qubits, 0., np.pi, "pool");
y_angles = rng.GetUniformRandomNumbers(num_qubits, 0., np.pi, "pool");
z_angles = rng.GetUniformRandomNumbers(num_qubits, 0., np.pi, "pool");
if False:
    info(x_angles)
    info(y_angles)
    info(z_angles)

# Ideal (i.e. noiseless) state.
# |psi> = |00000000>
psi = iqs.QubitRegister(num_qubits, "base", 0, 0);

# At this point we have one copy of the ideal state for each state in the pool.
info("\n---- ideal circuit \n")
for q in range(num_qubits):
    psi.ApplyRotationX (q, x_angles[q])
    psi.ApplyRotationY (q, y_angles[q])
    psi.ApplyRotationZ (q, z_angles[q])

iqs.MPIEnvironment.PoolBarrier()

# Compute the probability of qubit 0 to be in |1>.
probability = psi.GetProbability(0)

#################################################################################
# Quantum state evolution in presence of noise
#################################################################################

# State for slow decoherence.
psi_slow = iqs.QubitRegister(num_qubits, "base", 0, 0)
# One can use the same random number generator for each state or a different one.
# Here we use the same.
psi_slow.SetRngPtr(rng)
# T_1 and T_2 times for slow decoherence
T_1_slow , T_2_slow = 1000. , 500.
psi_slow.SetNoiseTimescales(T_1_slow, T_2_slow)

# State for fast decoherence.
psi_fast = iqs.QubitRegister(num_qubits, "base", 0, 0)
# Here too we use the same random number generator.
psi_fast.SetRngPtr(rng)
# T_1 and T_2 times for fast decoherence
T_1_fast , T_2_fast = 40. , 20.
psi_fast.SetNoiseTimescales(T_1_fast, T_2_fast)

# All single-qubit rotations have the same duration:
duration = 1.5
 
# ---------------- slow decoherence
info("---- slow decoherence \n")
iqs.MPIEnvironment.PoolBarrier()
overlap_squared_slow = 0
probability_slow = 0
for j in range(num_ensemble_states//num_pool_states):
    # Quantum circuit with explicit noise gates added to simulate noise.
    psi_slow.Initialize("base", 0)
    # Noise after state preparation:
    for q in range(num_qubits):
        psi_slow.ApplyNoiseGate (q, duration)
    # X-basis rotations.
    for qubit in range(num_qubits):
        psi_slow.ApplyRotationX (qubit, x_angles[qubit])
        # Since the gates are sequential, noise acts on all qubits after each gate.
        for q in range(num_qubits):
            psi_slow.ApplyNoiseGate (q, duration)
    # Y-basis rotations.
    for qubit in range(num_qubits):
        psi_slow.ApplyRotationY (qubit, y_angles[qubit])
        # Since the gates are sequential, noise acts on all qubits after each gate.
        for q in range(num_qubits):
            psi_slow.ApplyNoiseGate (q, duration)
    # Z-basis rotations.
    for qubit in range(num_qubits):
        psi_slow.ApplyRotationZ (qubit, z_angles[qubit])
        # Since the gates are sequential, noise acts on all qubits after each gate.
        for q in range(num_qubits):
            psi_slow.ApplyNoiseGate (q, duration)
    # Noise before state measurement has already been implemented.

    # Compute the probability of qubit 0 to be in |1>.
    probability_slow += psi_slow.GetProbability(0)

    # Overlap with ideal state.
    overlap_squared_slow += np.absolute( psi.ComputeOverlap(psi_slow) )**2

# Compute average per group in the pool.
overlap_squared_slow /= (num_ensemble_states/num_pool_states)
probability_slow /= (num_ensemble_states/num_pool_states)

# Incoherent average across the pool.
overlap_squared_slow = iqs.MPIEnvironment.IncoherentSumOverAllStatesOfPool(overlap_squared_slow)
overlap_squared_slow /= iqs.MPIEnvironment.GetNumStates()

probability_slow = iqs.MPIEnvironment.IncoherentSumOverAllStatesOfPool(probability_slow)
probability_slow /= iqs.MPIEnvironment.GetNumStates()

# NOTE: For the noise model considered, noise gates can be fused with each other.
#       In fact, the overall effect on the ensemble is the same for two consecutive
#       noise gates (on the same qubit) as a single one with duration equal to the
#       sum of the individual durations.
#       Exploiting this fact can reduce the computation time considerably.
#       In the example of this tutorial, instead of using:
#           N + 3*N*N 
#       noise gates one could have used just:
#           4*N
#       noise gates.

# ---------------- fast decoherence
info("---- fast decoherence \n")
iqs.MPIEnvironment.PoolBarrier()
overlap_squared_fast = 0
probability_fast = 0
for j in range(num_ensemble_states//num_pool_states):
    # Quantum circuit with explicit noise gates added to simulate noise.
    psi_fast.Initialize("base", 0)
    # Noise after state preparation:
    for q in range(num_qubits):
        psi_fast.ApplyNoiseGate (q, duration)
    # X-basis rotations.
    for qubit in range(num_qubits):
        psi_fast.ApplyRotationX (qubit, x_angles[qubit])
        # Since the gates are sequential, noise acts on all qubits after each gate.
        for q in range(num_qubits):
            psi_fast.ApplyNoiseGate (q, duration)
    # Y-basis rotations.
    for qubit in range(num_qubits):
        psi_fast.ApplyRotationY (qubit, y_angles[qubit])
        # Since the gates are sequential, noise acts on all qubits after each gate.
        for q in range(num_qubits):
            psi_fast.ApplyNoiseGate (q, duration)
    # Z-basis rotations.
    for qubit in range(num_qubits):
        psi_fast.ApplyRotationZ (qubit, z_angles[qubit])
        # Since the gates are sequential, noise acts on all qubits after each gate.
        for q in range(num_qubits):
            psi_fast.ApplyNoiseGate (q, duration)
    # Noise before state measurement has already been implemented.

    # Compute the probability of qubit 0 to be in |1>.
    probability_fast += psi_fast.GetProbability(0)

    # Overlap with ideal state.
    overlap_squared_fast += np.absolute( psi.ComputeOverlap(psi_fast) )**2

# Compute average per group in the pool.
overlap_squared_fast /= (num_ensemble_states/num_pool_states)
probability_fast /= (num_ensemble_states/num_pool_states)

# Incoherent average across the pool.
overlap_squared_fast = iqs.MPIEnvironment.IncoherentSumOverAllStatesOfPool(overlap_squared_fast)
overlap_squared_fast /= iqs.MPIEnvironment.GetNumStates()

probability_fast = iqs.MPIEnvironment.IncoherentSumOverAllStatesOfPool(probability_fast)
probability_fast /= iqs.MPIEnvironment.GetNumStates()

#################################################################################
# Quantum state initialization and ideal circuit
#################################################################################

# Print a few information on screen.
# Computation of the overlap between the ideal state and those exposed to noise:
info("---- summary of simulation:")
info("Overlap-squared between ideal and 'slow decoherence' state = {}".format(overlap_squared_slow))
info("Overlap-squared between ideal and 'fast decoherence' state = {}".format(overlap_squared_fast))
info("----")
info("Probability in the noiseless case = {}".format(probability))
info("Probability with slow decoherence = {}".format(probability_slow))
info("Probability with fast decoherence = {}\n".format(probability_fast))

# e = psi2.MaxAbsDiff(psi1);

# Finalize the MPI environment
iqs.EnvFinalize()



""" Qubit class definition """

from math import pi
from typing import TypeVar

QIntType = TypeVar("QIntType", bound="QUInt")
QubitType = TypeVar("QubitType", bound="Qubit")


class Qubit:
  """ A single qubit in a Q(U)Int """

  def __init__(self, quint: QIntType, qubit: int) -> None:
    if qubit >= quint.num_qubits:
      raise Exception("Qubit index out of bounds")

    self.quint = quint
    self.qubit = qubit

  def negate(self) -> QubitType:
    """ Negates the qubit (PauliX Gate)

    Returns:
      The Qubit

    """

    self.quint.negate(self.qubit)
    return self

  def c_negate(self, on: QubitType) -> QubitType:
    """ Controlled negate on another qubit

    Returns:
      The Qubit

    """

    self.quint.c_negate(on=on, qubit=self.qubit)
    return self

  def sqrt_not(self) -> QubitType:
    """ Apply square root of `not`. Applying this twice is a `not` gate

    Returns:
      The Qubit

    """

    self.quint.sqrt_not(self.qubit)
    return self

  def hadamard(self) -> QubitType:
    """ Applies a hadamard gate. Puts the qubit in an exact superposition

    Returns:
      The Qubit

    """

    self.quint.hadamard(self.qubit)
    return self

  def phase(self, angle: float = pi / 2) -> QubitType:
    """ Applies a hadamard gate. Puts the qubit in an exact superposition

    Returns:
      The Qubit

    """

    self.quint.phase(self.qubit, angle)
    return self

  def c_phase(self, *, on: QubitType, angle: float = pi / 2) -> QubitType:
    """ Applies a hadamard gate. Puts the qubit in an exact superposition

    Returns:
      The Qubit

    """

    self.quint.c_phase(on=on, qubit=self.qubit, angle=angle)
    return self

  def swap(self, other: QubitType) -> QubitType:
    """ Swap with a give qubit

    Args:
      other: Qubit to swap with

    Returns:
      The original Qubit method was called on

    """

    self.quint.swap(other.quint, self.qubit, other.qubit)
    return self

  def c_swap(self, other: QubitType, on: QubitType) -> QubitType:
    self.quint.c_swap(other.quint, self.qubit, other.qubit, on=on)
    return self

  def measure(self) -> int:
    return self.quint.measure(self.qubit)

  # Aliases
  h = hadamard
  dburby = hadamard
  entangle = c_negate



""" Internal IQS wrapper """

from os import path
import sys

# pylint: disable-next=wrong-import-position
repo_root = path.dirname(path.dirname(path.abspath(__file__)))
lib_path = path.join(repo_root, "intel-qs", "build", "lib")
sys.path.insert(0, lib_path)
import intelqs_py as _iqs


def cpp_method(*args, **kwargs):
  raise Exception("Method only exists in C++. See qublets/iqs.py to enable")


# List of methods in C++ that are not exposed to Python.
# If you'd like to use any, add the given config line to
# intel-qs/pybind11/intelqs_py.cpp locally and rebuild.
# TODO(dyordan1): Push this upstream into intel-qs

# In py::class_< QubitRegister<ComplexDP> >:
# .def("GetGlobalAmplitude", &QubitRegister<ComplexDP>::GetGlobalAmplitude)
if 'GetGlobalAmplitude' not in dir(_iqs.QubitRegister):
  _iqs.QubitRegister.GetGlobalAmplitude = cpp_method

# .def("IsClassicalBit", &QubitRegister<ComplexDP>::IsClassicalBit)
if 'IsClassicalBit' not in dir(_iqs.QubitRegister):
  _iqs.QubitRegister.IsClassicalBit = cpp_method


""" Simulation class definition """

import time
from typing import Callable, List, Tuple


class Simulation:
  """ Utility class for running simulations in batches

  Args:
    num_shots: How many times to run a simulation for
    function: The simulation function to run

  Attributes:
    last_results: The results (distribution) for the most-recently-ran batch
    function: The elapsed time (ms) for the most-recently-ran batch

  """

  def __init__(self, num_shots: int, function: Callable[[], int]) -> None:
    self.num_shots = num_shots
    self.function = function
    self.last_results = None
    self.last_time = None

  def run(self) -> List[Tuple[int, int]]:
    """ Run the simulation, returning the distribution of results

    Returns:
      A sorted tuple reprenting a histogram of values returned by the
      simulation function

    """

    start = time.time()
    results = {}
    for _ in range(self.num_shots):
      result = self.function()
      results.setdefault(result, 0)
      results[result] += 1

    self.last_results = sorted(results.items(), key=lambda x: x[0])
    self.last_time = round((time.time() - start) * 1000, 2)

    return self.last_results



""" Pip package setup """

import glob
from os import path, makedirs, environ
import pybind11
from shutil import copytree, rmtree
import subprocess
import sys

from setuptools import Extension, setup
from setuptools.command.build_ext import build_ext


class CMakeExtension(Extension):

  def __init__(self, name, sources):
    Extension.__init__(self, name, sources=sources)


class CMakeBuild(build_ext):

  def build_extension(self, ext):
    extdir = path.abspath(path.dirname(self.get_ext_fullpath(ext.name)))
    extdir = path.join(extdir, "intel-qs", "build", "lib")

    if not path.exists(self.build_temp):
      makedirs(self.build_temp)

    cmake_command = [
        "cmake", f"-DPYTHON_EXECUTABLE={sys.executable}",
        f"-DCMAKE_PREFIX_PATH={pybind11.get_cmake_dir()}", "-DIqsPython=ON",
        "-DIqsNative=ON", "../../intel-qs"
    ]
    print(cmake_command)
    env = environ.copy()
    env["CXX"] = "g++"
    subprocess.check_call(cmake_command, cwd=self.build_temp, env=env)
    subprocess.check_call(["make", "-j", "8"], cwd=self.build_temp, env=env)

    rmtree(extdir, ignore_errors=True)
    copytree(path.join(self.build_temp, "lib"), extdir)


cmake_dir = glob.glob(path.join("intel-qs", "cmake", "**"), recursive=True)
src_dir = glob.glob(path.join("intel-qs", "src", "**"), recursive=True)
inc_dir = glob.glob(path.join("intel-qs", "include", "**"), recursive=True)
pybind_dir = glob.glob(path.join("intel-qs", "pybind11", "**"), recursive=True)
tutorials_dir = glob.glob(path.join("intel-qs", "tutorials", "**"),
                          recursive=True)
benchmarks_dir = glob.glob(path.join("intel-qs", "benchmarks", "**"),
                           recursive=True)
other = [
    path.join("intel-qs", "CMakeLists.txt"),
    path.join("intel-qs", "LICENSE.md")
]
iqs_files = cmake_dir + src_dir + inc_dir + pybind_dir + tutorials_dir + benchmarks_dir + other

long_description = ""
with open("README.md") as f:
  long_description = f.read()

setup(name="qublets",
      version="0.1.3",
      description="A quantum computing library for the rest of us",
      author="Dobromir Yordanov",
      author_email="dobri.domain@gmail.com",
      long_description=long_description,
      long_description_content_type="text/markdown",
      license="",
      url="https://github.com/dyordan1/qublets",
      packages=["qublets"],
      ext_modules=[CMakeExtension("qublets", iqs_files)],
      cmdclass={"build_ext": CMakeBuild},
      python_requires=">=3.6",
      install_requires=["dataclasses"])
